// TestSharedPointerRacecondition.cpp
//
// c++ shared pointers are *not* thread-safe. We simulate this by
// creating enough threads and introducing enough latency.
//
// Inside every shared pointer there is a global/singleton object
// trying to get it. It goes unnoticed in a single-threaded application.
// It could even go unnoticed in multi-threaded code.
//
// (C) Datasim Education BV 2015
//

#include <thread>
#include <atomic>
#include <memory>
#include <random>
#include <functional>
#include <mutex>

#include <iostream>

struct X
{
	double val;

	X() : val(0.0) {}
	void operator ()()
	{
		std::cout << "An X " << val << std::endl;
	}
};


// C++11
template <typename T>
	using GenericPointerType = std::shared_ptr<T>;
using PointerType = GenericPointerType<X>;

// Function to test race condition
std::mutex displayMutex;
int latencyWait = 50'000; // ms

void Modify(PointerType& p, double newVal)
{
	// Wait a while, long enough to trigger the race
	std::random_device rd; // Slows things down

	std::default_random_engine dre(rd());
	std::uniform_int_distribution<int> delay(0, latencyWait);
	std::this_thread::sleep_for(std::chrono::milliseconds(delay(dre)));

	p->val = newVal;

	displayMutex.lock();
		std::cout << "Thread ID, std::thread::id " << std::this_thread::get_id() << std::endl;
	displayMutex.unlock();

}

int main()
{
		

			// Counter-example to show race conditions with shared pointers
			std::cout << "Smart races, processing...\n";
			PointerType x(new X);
			x->val = 2.71;

			double d1 = 1.0;
			double d2 = 2.0;
			double d3 = 3.0;
			double d4 = 4.0;

			std::thread t4(Modify, x, d1);
			std::thread t2(Modify, x, d2);
			std::thread t3(Modify, x, d3);
			std::thread t1(Modify, x, d4); 

			t1.join();
			t2.join();
			t3.join();
			t4.join();

			std::cout << "Value: " << x->val << std::endl;

			// Now swamp with *many* threads
			int numThreads = 100;
			latencyWait = 2000; 
			std::vector<std::thread> v;
			for (int n = 0; n < numThreads; ++n)
			{
				v.emplace_back(std::bind(Modify, x, static_cast<double>(n)));
			}
			for (auto& t : v) 
			{
				t.join();
			}

			std::cout << "Value: " << x->val << std::endl;

			return 0;
}
